@@ -26,12 +26,16 @@
 #ifndef _COMPIZ_H
 #define _COMPIZ_H
 
+#define ABIVERSION 20060927
+
 #include <stdio.h>
 #include <sys/time.h>
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/extensions/Xdamage.h>
+#include <X11/extensions/Xcomposite.h>
+#include <X11/extensions/Xinerama.h>
 #include <X11/extensions/sync.h>
 #include <X11/Xregion.h>
 #include <X11/XKBlib.h>
@@ -42,6 +46,10 @@
 #include <GL/gl.h>
 #include <GL/glx.h>
 
+#if COMPOSITE_MAJOR > 0 || COMPOSITE_MINOR > 2
+#define USE_COW
+#endif
+
 /*
  * WORDS_BIGENDIAN should be defined before including this file for
  * IMAGE_BYTE_ORDER and BITMAP_BIT_ORDER to be set correctly.
@@ -194,6 +202,7 @@ extern CompWindow *lastDamagedWindow;
 extern Bool       replaceCurrentWm;
 extern Bool       indirectRendering;
 extern Bool       strictBinding;
+extern Bool       useCow;
 
 extern int  defaultRefreshRate;
 extern char *defaultTextureFilter;
@@ -508,8 +517,7 @@ typedef int CompWatchFdHandle;
 #define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT  51
 #define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS     52
 #define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED          53
-#define COMP_DISPLAY_OPTION_RAISE_WINDOW                  54
-#define COMP_DISPLAY_OPTION_NUM                           55
+#define COMP_DISPLAY_OPTION_NUM                           54
 
 typedef CompOption *(*GetDisplayOptionsProc) (CompDisplay *display,
 					      int	  *count);
@@ -553,6 +561,12 @@ struct _CompDisplay {
     Bool xkbExtension;
     int  xkbEvent, xkbError;
 
+    Bool xineramaExtension;
+    int  xineramaEvent, xineramaError;
+
+    XineramaScreenInfo *screenInfo;
+    int		       nScreenInfo;
+
     SnDisplay *snDisplay;
 
     Atom supportedAtom;
@@ -679,8 +693,6 @@ struct _CompDisplay {
 
     GLenum textureFilter;
 
-    Window glIncludeInferiorsOwner;
-
     Window activeWindow;
 
     Window below;
@@ -753,6 +765,9 @@ addDisplay (char *name,
 	    char **plugin,
 	    int  nPlugin);
 
+Time
+getCurrentTimeFromDisplay (CompDisplay *d);
+
 void
 focusDefaultWindow (CompDisplay *d);
 
@@ -883,10 +898,12 @@ typedef void (*DonePaintScreenProc) (CompScreen *screen);
 typedef Bool (*PaintScreenProc) (CompScreen		 *screen,
 				 const ScreenPaintAttrib *sAttrib,
 				 Region			 region,
+				 int		         output,
 				 unsigned int		 mask);
 
 typedef void (*PaintTransformedScreenProc) (CompScreen		    *screen,
 					    const ScreenPaintAttrib *sAttrib,
+					    int			    output,
 					    unsigned int	    mask);
 
 
@@ -894,7 +911,6 @@ typedef void (*PaintTransformedScreenProc) (CompScreen		    *screen,
 #define PAINT_WINDOW_TRANSLUCENT_MASK		(1 << 1)
 #define PAINT_WINDOW_TRANSFORMED_MASK           (1 << 2)
 #define PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK (1 << 3)
-#define PAINT_WINDOW_DECORATION_MASK            (1 << 4)
 
 typedef Bool (*PaintWindowProc) (CompWindow		 *window,
 				 const WindowPaintAttrib *attrib,
@@ -935,12 +951,14 @@ translateRotateScreen (const ScreenPaintAttrib *sa);
 void
 paintTransformedScreen (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
+			int			output,
 			unsigned int	        mask);
 
 Bool
 paintScreen (CompScreen		     *screen,
 	     const ScreenPaintAttrib *sAttrib,
 	     Region		     region,
+	     int		     output,
 	     unsigned int	     mask);
 
 Bool
@@ -1063,11 +1081,13 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_DETECT_REFRESH_RATE 0
 #define COMP_SCREEN_OPTION_LIGHTING	       1
 #define COMP_SCREEN_OPTION_REFRESH_RATE	       2
-#define COMP_SCREEN_OPTION_SIZE		       3
-#define COMP_SCREEN_OPTION_OPACITY_STEP        4
-#define COMP_SCREEN_OPTION_UNREDIRECT_FS       5
-#define COMP_SCREEN_OPTION_DEFAULT_ICON        6
-#define COMP_SCREEN_OPTION_NUM		       7
+#define COMP_SCREEN_OPTION_HSIZE	       3
+#define COMP_SCREEN_OPTION_VSIZE	       4
+#define COMP_SCREEN_OPTION_OPACITY_STEP        5
+#define COMP_SCREEN_OPTION_UNREDIRECT_FS       6
+#define COMP_SCREEN_OPTION_DEFAULT_ICON        7
+#define COMP_SCREEN_OPTION_SYNC_TO_VBLANK      8
+#define COMP_SCREEN_OPTION_NUM		       9
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1112,6 +1132,11 @@ typedef void (*GLXCopySubBufferProc) (Display     *display,
 				      int	  width,
 				      int	  height);
 
+typedef int (*GLXGetVideoSyncProc)  (unsigned int *count);
+typedef int (*GLXWaitVideoSyncProc) (int	  divisor,
+				     int	  remainder,
+				     unsigned int *count);
+
 #ifndef GLX_VERSION_1_3
 typedef struct __GLXFBConfigRec *GLXFBConfig;
 #endif
@@ -1244,7 +1269,8 @@ typedef struct _CompGroup {
 typedef struct _CompStartupSequence {
     struct _CompStartupSequence *next;
     SnStartupSequence		*sequence;
-    unsigned int		viewport;
+    unsigned int		viewportX;
+    unsigned int		viewportY;
 } CompStartupSequence;
 
 typedef struct _CompFBConfig {
@@ -1273,16 +1299,17 @@ typedef struct _CompScreenEdge {
     unsigned int count;
 } CompScreenEdge;
 
-typedef struct _CompScreenHead {
-    int x, y, width, height, areaX, areaY, areaWidth, areaHeight;
-} CompScreenHead;
-
 struct _CompIcon {
     CompTexture texture;
     int		width;
     int		height;
 };
 
+typedef struct _CompOutput {
+    char   *name;
+    REGION region;
+} CompOutput;
+
 struct _CompScreen {
     CompScreen  *next;
     CompDisplay *display;
@@ -1297,11 +1324,15 @@ struct _CompScreen {
     int		      width;
     int		      height;
     int		      x;
-    int		      size;
+    int		      y;
+    int		      hsize;		/* Number of horizontal viewports */
+    int		      vsize;		/* Number of vertical viewports */
     REGION	      region;
     Region	      damage;
     unsigned long     damageMask;
     Window	      root;
+    Window	      overlay;
+    Window	      output;
     XWindowAttributes attrib;
     Window	      grabWindow;
     CompFBConfig      glxPixmapFBConfigs[MAX_DEPTH + 1];
@@ -1324,6 +1355,10 @@ struct _CompScreen {
     unsigned int      mapNum;
     unsigned int      activeNum;
 
+    CompOutput *outputDev;
+    int	       nOutputDev;
+    int	       currentOutputDev;
+
     int overlayWindowCount;
 
     CompScreenEdge screenEdge[SCREEN_EDGE_NUM];
@@ -1388,6 +1423,8 @@ struct _CompScreen {
     GLXReleaseTexImageProc   releaseTexImage;
     GLXQueryDrawableProc     queryDrawable;
     GLXCopySubBufferProc     copySubBuffer;
+    GLXGetVideoSyncProc      getVideoSync;
+    GLXWaitVideoSyncProc     waitVideoSync;
     GLXGetFBConfigsProc      getFBConfigs;
     GLXGetFBConfigAttribProc getFBConfigAttrib;
     GLXCreatePixmapProc      createPixmap;
@@ -1437,11 +1474,6 @@ struct _CompScreen {
     WindowUngrabNotifyProc windowUngrabNotify;
 
     CompPrivate *privates;
-
-    Bool 	   useXinerama;
-    int		   headsCount;
-
-    CompScreenHead *headsInfo;
 };
 
 int
@@ -1459,6 +1491,10 @@ void
 configureScreen (CompScreen	 *s,
 		 XConfigureEvent *ce);
 
+void
+setCurrentOutput (CompScreen *s,
+		  int	     outputNum);
+
 void
 updateScreenBackground (CompScreen  *screen,
 			CompTexture *texture);
@@ -1466,39 +1502,6 @@ updateScreenBackground (CompScreen  *screen,
 void
 detectRefreshRateOfScreen (CompScreen *s);
 
-int 
-screenGetHead ( CompScreen *s, int x, int y, int w, int h );
-
-int
-screenGetHeadForWindow ( CompWindow *w );
-
-int
-screenGetCurrentHead ( CompScreen *s );
-
-int
-screenGetHeadX ( CompScreen *s, int head );
-
-int
-screenGetHeadY ( CompScreen *s, int head );
-
-int
-screenGetHeadWidth ( CompScreen *s, int head );
-
-int
-screenGetHeadHeight ( CompScreen *s, int head );
-
-int
-screenGetHeadAreaX ( CompScreen *s, int head );
-
-int
-screenGetHeadAreaY ( CompScreen *s, int head );
-
-int
-screenGetHeadAreaWidth ( CompScreen *s, int head );
-
-int
-screenGetHeadAreaHeight ( CompScreen *s, int head );
-
 Bool
 addScreen (CompDisplay *display,
 	   int	       screenNum,
@@ -1534,11 +1537,6 @@ CompWindow *
 findWindowAtScreen (CompScreen *s,
 		    Window     id);
 
-Bool
-pointerOnlyOnDesktop (CompScreen *s,
-                      int pointerX,
-                      int pointerY);
-
 CompWindow *
 findTopLevelWindowAtScreen (CompScreen *s,
 			    Window      id);
@@ -1598,6 +1596,7 @@ runCommand (CompScreen *s,
 void
 moveScreenViewport (CompScreen *s,
 		    int	       tx,
+		    int	       ty,
 		    Bool       sync);
 
 void
@@ -1631,10 +1630,6 @@ void
 sendWindowActivationRequest (CompScreen *s,
 			     Window	id);
 
-void
-changeToWindowViewport (CompScreen *s,
-			     Window	id, CompWindow *w);
-
 void
 screenTexEnvMode (CompScreen *s,
 		  GLenum     mode);
@@ -1660,6 +1655,11 @@ makeScreenCurrent (CompScreen *s);
 void
 finishScreenDrawing (CompScreen *s);
 
+int
+outputDeviceForPoint (CompScreen *s,
+		      int	 x,
+		      int	 y);
+
 
 /* window.c */
 
@@ -1731,7 +1731,8 @@ struct _CompWindow {
     int		      destroyRefCnt;
     int		      unmapRefCnt;
 
-    unsigned int initialViewport;
+    unsigned int initialViewportX;
+    unsigned int initialViewportY;
 
     Bool placed;
     Bool minimized;
@@ -2023,9 +2024,6 @@ updateWindowAttributes (CompWindow *w,
 void
 activateWindow (CompWindow *w);
 
-void
-activateWindowLight (CompWindow *w);
-
 void
 closeWindow (CompWindow *w,
 	     Time	serverTime);
@@ -2074,8 +2072,10 @@ unredirectWindow (CompWindow *w);
 void
 redirectWindow (CompWindow *w);
 
-int
-defaultViewportForWindow (CompWindow *w);
+void
+defaultViewportForWindow (CompWindow *w,
+			  int	     *vx,
+			  int        *vy);
 
 CompIcon *
 getWindowIcon (CompWindow *w,
@@ -2085,9 +2085,15 @@ getWindowIcon (CompWindow *w,
 void
 freeWindowIcons (CompWindow *w);
 
+int
+outputDeviceForWindow (CompWindow *w);
+
 
 /* plugin.c */
 
+typedef int (*GetVersionProc) (CompPlugin *plugin,
+			       int	  version);
+
 typedef Bool (*InitPluginProc) (CompPlugin *plugin);
 typedef void (*FiniPluginProc) (CompPlugin *plugin);
 
@@ -2106,6 +2112,8 @@ typedef struct _CompPluginVTable {
     char *shortDesc;
     char *longDesc;
 
+    GetVersionProc getVersion;
+
     InitPluginProc init;
     FiniPluginProc fini;
 
@@ -2129,6 +2137,15 @@ typedef struct _CompPluginVTable {
 
 typedef CompPluginVTable *(*PluginGetInfoProc) (void);
 
+typedef Bool (*LoadPluginProc) (CompPlugin *p,
+				char	   *path,
+				char	   *name);
+
+typedef void (*UnloadPluginProc) (CompPlugin *p);
+
+extern LoadPluginProc   loaderLoadPlugin;
+extern UnloadPluginProc loaderUnloadPlugin;
+
 struct _CompPlugin {
     CompPlugin       *next;
     CompPrivate	     devPrivate;
@@ -2136,15 +2153,6 @@ struct _CompPlugin {
     CompPluginVTable *vTable;
 };
 
-typedef Bool (*LoadPluginProc) (CompPlugin *p,
-				char	   *path,
-				char	   *name);
- 
-typedef void (*UnloadPluginProc) (CompPlugin *p);
-
-extern LoadPluginProc   loaderLoadPlugin;
-extern UnloadPluginProc loaderUnloadPlugin;
-
 Bool
 initPluginForDisplay (CompPlugin  *p,
 		      CompDisplay *d);
