@@ -81,6 +81,10 @@ Bool replaceCurrentWm = FALSE;
 Bool indirectRendering = FALSE;
 Bool strictBinding = FALSE;
 
+#ifdef USE_COW
+Bool useCow = FALSE;
+#endif
+
 static void
 usage (void)
 {
@@ -96,6 +100,11 @@ usage (void)
 	    "[--sm-disable] "
 	    "[--sm-client-id ID] "
 	    "[--version]\n       "
+
+#ifdef USE_COW
+	    "[--use-cow] "
+#endif
+
 	    "[--help] "
 	    "[PLUGIN]...\n",
 	    programName);
@@ -186,6 +195,14 @@ main (int argc, char **argv)
 	{
 	    strictBinding = TRUE;
 	}
+
+#ifdef USE_COW
+	else if (!strcmp (argv[i], "--use-cow"))
+	{
+	    useCow = TRUE;
+	}
+#endif
+
 	else if (!strcmp (argv[i], "--replace"))
 	{
 	    replaceCurrentWm = TRUE;
