@@ -40,17 +40,20 @@
 #include <X11/Xatom.h>
 #include <X11/Xproto.h>
 #include <X11/extensions/Xrandr.h>
+#include <X11/extensions/shape.h>
 #include <X11/cursorfont.h>
 
-#include <X11/extensions/Xinerama.h>
-
 #include <compiz.h>
 
 #define DETECT_REFRESH_RATE_DEFAULT TRUE
 
-#define SCREEN_SIZE_DEFAULT 4
-#define SCREEN_SIZE_MIN	    4
-#define SCREEN_SIZE_MAX	    32
+#define SCREEN_HSIZE_DEFAULT 4
+#define SCREEN_HSIZE_MIN     1
+#define SCREEN_HSIZE_MAX     32
+
+#define SCREEN_VSIZE_DEFAULT 1
+#define SCREEN_VSIZE_MIN     1
+#define SCREEN_VSIZE_MAX     32
 
 #define LIGHTING_DEFAULT TRUE
 
@@ -62,6 +65,8 @@
 
 #define DEFAULT_ICON_DEFAULT "icon.png"
 
+#define SYNC_TO_VBLANK_DEFAULT TRUE
+
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
 static int
@@ -104,19 +109,21 @@ freeScreenPrivateIndex (CompDisplay *display,
 
 static void
 setVirtualScreenSize (CompScreen *screen,
-		      int	 size)
+		      int	 hsize,
+		      int	 vsize)
 {
     unsigned long data[2];
 
-    data[0] = screen->width * size;
-    data[1] = screen->height;
+    data[0] = screen->width * hsize;
+    data[1] = screen->height * vsize;
 
     XChangeProperty (screen->display->display, screen->root,
 		     screen->display->desktopGeometryAtom,
 		     XA_CARDINAL, 32, PropModeReplace,
 		     (unsigned char *) data, 2);
 
-    screen->size = size;
+    screen->hsize = hsize;
+    screen->vsize = vsize;
 }
 
 static Bool
@@ -181,6 +188,7 @@ setScreenOption (CompScreen      *screen,
     case COMP_SCREEN_OPTION_DETECT_REFRESH_RATE:
     case COMP_SCREEN_OPTION_LIGHTING:
     case COMP_SCREEN_OPTION_UNREDIRECT_FS:
+    case COMP_SCREEN_OPTION_SYNC_TO_VBLANK:
 	if (compSetBoolOption (o, value))
 	    return TRUE;
 	break;
@@ -195,13 +203,27 @@ setScreenOption (CompScreen      *screen,
 	    return TRUE;
 	}
 	break;
-    case COMP_SCREEN_OPTION_SIZE:
+    case COMP_SCREEN_OPTION_HSIZE:
 	if (compSetIntOption (o, value))
 	{
+	    CompOption *vsize = compFindOption (screen->opt, NUM_OPTIONS (screen), "vsize", NULL);
+	    
 	    if (o->value.i * screen->width > MAXSHORT)
 		return FALSE;
 
-	    setVirtualScreenSize (screen, o->value.i);
+	    setVirtualScreenSize (screen, o->value.i, vsize->value.i);
+	    return TRUE;
+	}
+	break;
+    case COMP_SCREEN_OPTION_VSIZE:
+	if (compSetIntOption (o, value))
+	{
+	    CompOption *hsize = compFindOption (screen->opt, NUM_OPTIONS (screen), "hsize", NULL);
+
+	    if (o->value.i * screen->height > MAXSHORT)
+		return FALSE;
+
+	    setVirtualScreenSize (screen, hsize->value.i, o->value.i);
 	    return TRUE;
 	}
 	break;
@@ -265,14 +287,23 @@ compScreenInitOptions (CompScreen *screen)
     o->rest.i.min = 1;
     o->rest.i.max = 200;
 
-    o = &screen->opt[COMP_SCREEN_OPTION_SIZE];
-    o->name	  = "size";
-    o->shortDesc  = N_("Virtual Size");
-    o->longDesc	  = N_("Screen size multiplier for virtual size");
+    o = &screen->opt[COMP_SCREEN_OPTION_HSIZE];
+    o->name	  = "hsize";
+    o->shortDesc  = N_("Horizontal Virtual Size");
+    o->longDesc	  = N_("Screen size multiplier for horizontal virtual size");
     o->type	  = CompOptionTypeInt;
-    o->value.i    = SCREEN_SIZE_DEFAULT;
-    o->rest.i.min = SCREEN_SIZE_MIN;
-    o->rest.i.max = SCREEN_SIZE_MAX;
+    o->value.i    = SCREEN_HSIZE_DEFAULT;
+    o->rest.i.min = SCREEN_HSIZE_MIN;
+    o->rest.i.max = SCREEN_HSIZE_MAX;
+
+    o = &(screen->opt[COMP_SCREEN_OPTION_VSIZE]);
+    o->name	      = "vsize";
+    o->shortDesc      = N_("Vertical Virtual Size");
+    o->longDesc	      = N_("Screen size multiplier for vertical virtual size");
+    o->type	      = CompOptionTypeInt;
+    o->value.i	      = SCREEN_VSIZE_DEFAULT;
+    o->rest.i.min     = SCREEN_VSIZE_MIN;
+    o->rest.i.max     = SCREEN_VSIZE_MAX;
 
     o = &screen->opt[COMP_SCREEN_OPTION_OPACITY_STEP];
     o->name		= "opacity_step";
@@ -299,6 +330,14 @@ compScreenInitOptions (CompScreen *screen)
     o->value.s	      = strdup (DEFAULT_ICON_DEFAULT);
     o->rest.s.string  = 0;
     o->rest.s.nString = 0;
+
+    o = &screen->opt[COMP_SCREEN_OPTION_SYNC_TO_VBLANK];
+    o->name       = "sync_to_vblank";
+    o->shortDesc  = N_("Sync To VBlank");
+    o->longDesc   = N_("Only perform screen updates during vertical "
+		       "blanking period");
+    o->type       = CompOptionTypeBool;
+    o->value.b    = SYNC_TO_VBLANK_DEFAULT;
 }
 
 static void
@@ -352,7 +391,8 @@ addSequence (CompScreen        *screen,
 
     s->next     = screen->startupSequences;
     s->sequence = sequence;
-    s->viewport = screen->x;
+    s->viewportX = screen->x;
+    s->viewportY = screen->y;
 
     screen->startupSequences = s;
 
@@ -496,14 +536,87 @@ perspective (GLfloat fovy,
    frustum (xmin, xmax, ymin, ymax, zNear, zFar);
 }
 
+static void
+updateOutputDevices (CompScreen *s)
+{
+    CompOutput *output;
+    int	       nOutput, i;
+
+    if (s->display->screenInfo)
+    {
+	XineramaScreenInfo *screenInfo = s->display->screenInfo;
+	int		   nScreenInfo = s->display->nScreenInfo;
+
+	output = malloc (sizeof (CompOutput) * nScreenInfo);
+	if (!output)
+	    return;
+
+	for (i = 0; i < nScreenInfo; i++)
+	{
+	    output->name = malloc (sizeof (char) * 10);
+	    if (output->name)
+		snprintf (output->name, 10, "Output %d", i);
+
+	    output[i].region.rects    = &output[i].region.extents;
+	    output[i].region.numRects = 1;
+
+	    output[i].region.extents.x1 = screenInfo[i].x_org;
+	    output[i].region.extents.y1 = screenInfo[i].y_org;
+	    output[i].region.extents.x2 = screenInfo[i].x_org +
+		screenInfo[i].width;
+	    output[i].region.extents.y2 = screenInfo[i].y_org +
+		screenInfo[i].height;
+	}
+
+	nOutput = nScreenInfo;
+    }
+    else
+    {
+	output = malloc (sizeof (CompOutput));
+	if (!output)
+	    return;
+
+	output->name   = strdup ("Output 0");
+
+	output->region.rects    = &output->region.extents;
+	output->region.numRects = 1;
+
+	output->region.extents.x1 = 0;
+	output->region.extents.y1 = 0;
+	output->region.extents.x2 = s->attrib.width;
+	output->region.extents.y2 = s->attrib.height;
+
+	nOutput = 1;
+    }
+
+    if (s->outputDev)
+    {
+	for (i = 0; i < s->nOutputDev; i++)
+	    if (s->outputDev->name)
+		free (s->outputDev->name);
+
+	free (s->outputDev);
+    }
+
+    s->outputDev  = output;
+    s->nOutputDev = nOutput;
+}
+
+void
+setCurrentOutput (CompScreen *s,
+		  int	     outputNum)
+{
+    if (outputNum >= s->nOutputDev)
+	outputNum = 0;
+
+    s->currentOutputDev = outputNum;
+}
+
 static void
 reshape (CompScreen *s,
 	 int	    w,
 	 int	    h)
 {
-    s->width  = w;
-    s->height = h;
-
     glMatrixMode (GL_PROJECTION);
     glLoadIdentity ();
     glMatrixMode (GL_MODELVIEW);
@@ -528,6 +641,13 @@ reshape (CompScreen *s,
     s->region.extents.y2 = h;
     s->region.size = 1;
 
+    s->width  = w;
+    s->height = h;
+
+    updateOutputDevices (s);
+
+    setCurrentOutput (s, s->currentOutputDev);
+
     updateScreenEdges (s);
     updateWorkareaForScreen (s);
 }
@@ -826,22 +946,25 @@ setDesktopHints (CompScreen *s)
 	{
 	    memcpy (data, propData, sizeof (unsigned long));
 
-	    if (data[0] / s->width < s->size - 1)
+	    if (data[0] / s->width < s->hsize - 1)
 		s->x = data[0] / s->width;
+
+	    if (data[1] / s->height < s->vsize - 1)
+		s->y = data[1] / s->height;
 	}
 
 	XFree (propData);
     }
 
     data[0] = s->x * s->width;
-    data[1] = 0;
+    data[1] = s->y * s->height;
 
     XChangeProperty (d->display, s->root, d->desktopViewportAtom,
 		     XA_CARDINAL, 32, PropModeReplace,
 		     (unsigned char *) data, 2);
 
-    data[0] = s->width * s->size;
-    data[1] = s->height;
+    data[0] = s->width * s->hsize;
+    data[1] = s->height * s->vsize;
 
     XChangeProperty (d->display, s->root, d->desktopGeometryAtom,
 		     XA_CARDINAL, 32, PropModeReplace,
@@ -865,7 +988,7 @@ setDesktopHints (CompScreen *s)
 				 d->showingDesktopAtom, 0L, 1L, FALSE,
 				 XA_CARDINAL, &actual, &format,
 				 &n, &left, &propData);
-
+    
     if (result == Success && n && propData)
     {
 	memcpy (data, propData, sizeof (unsigned long));
@@ -880,6 +1003,27 @@ setDesktopHints (CompScreen *s)
 		     (unsigned char *) data, 1);
 }
 
+#ifdef USE_COW
+static void
+makeOutputWindow (CompScreen *s)
+{
+    Display       *dpy = s->display->display;
+    XserverRegion region;
+
+    s->overlay   = XCompositeGetOverlayWindow (dpy, s->root);
+    s->output    = s->overlay;
+
+    region = XFixesCreateRegion (dpy, NULL, 0);
+
+    XFixesSetWindowShapeRegion (dpy,
+				s->output,
+				ShapeInput,
+				0, 0, region);
+
+    XFixesDestroyRegion (dpy, region);
+}
+#endif
+
 Bool
 addScreen (CompDisplay *display,
 	   int	       screenNum,
@@ -937,8 +1081,10 @@ addScreen (CompDisplay *display,
     if (!s->damage)
 	return FALSE;
 
-    s->x    = 0;
-    s->size = 4;
+    s->x     = 0;
+    s->y     = 0;
+    s->hsize = SCREEN_HSIZE_DEFAULT;
+    s->vsize = SCREEN_VSIZE_DEFAULT;
 
     for (i = 0; i < SCREEN_EDGE_NUM; i++)
     {
@@ -990,6 +1136,10 @@ addScreen (CompDisplay *display,
     s->rasterX = 0;
     s->rasterY = 0;
 
+    s->outputDev	= NULL;
+    s->nOutputDev	= 0;
+    s->currentOutputDev = 0;
+
     s->windows = 0;
     s->reverseWindows = 0;
 
@@ -1047,6 +1197,14 @@ addScreen (CompDisplay *display,
 
     s->grabWindow = None;
 
+#ifdef USE_COW
+    if (useCow)
+	makeOutputWindow (s);
+    else
+#endif
+
+	s->overlay = s->output = s->root;
+
     templ.visualid = XVisualIDFromVisual (s->attrib.visual);
 
     visinfo = XGetVisualInfo (dpy, VisualIDMask, &templ, &nvisinfo);
@@ -1115,12 +1273,7 @@ addScreen (CompDisplay *display,
 
     XFree (visinfo);
 
-    if (!indirectRendering)
-    	glxExtensions = glXQueryExtensionsString (s->display->display, 
-						screenNum);
-    else
-    	glxExtensions = glXQueryServerString (s->display->display, screenNum,
-			    			GLX_EXTENSIONS);
+    glxExtensions = glXQueryExtensionsString (s->display->display, screenNum);
     if (!strstr (glxExtensions, "GLX_EXT_texture_from_pixmap"))
     {
 	fprintf (stderr, "%s: GLX_EXT_texture_from_pixmap is missing\n",
@@ -1177,7 +1330,18 @@ addScreen (CompDisplay *display,
 	s->copySubBuffer = (GLXCopySubBufferProc)
 	    getProcAddress (s, "glXCopySubBufferMESA");
 
-    glXMakeCurrent (dpy, s->root, s->ctx);
+    s->getVideoSync = NULL;
+    s->waitVideoSync = NULL;
+    if (strstr (glxExtensions, "GLX_SGI_video_sync"))
+    {
+	s->getVideoSync = (GLXGetVideoSyncProc)
+	    getProcAddress (s, "glXGetVideoSyncSGI");
+
+	s->waitVideoSync = (GLXWaitVideoSyncProc)
+	    getProcAddress (s, "glXWaitVideoSyncSGI");
+    }
+
+    glXMakeCurrent (dpy, s->output, s->ctx);
     currentRoot = s->root;
 
     glExtensions = (const char *) glGetString (GL_EXTENSIONS);
@@ -1200,7 +1364,7 @@ addScreen (CompDisplay *display,
 	    GLint maxTextureSize;
 
 	    glGetIntegerv (GL_MAX_RECTANGLE_TEXTURE_SIZE_NV, &maxTextureSize);
-	    if (maxTextureSize < s->maxTextureSize)
+	    if (maxTextureSize > s->maxTextureSize)
 		s->maxTextureSize = maxTextureSize;
 	}
     }
@@ -1309,15 +1473,19 @@ addScreen (CompDisplay *display,
 
 	for (j = 0; j < nElements; j++)
 	{
-	    {
-	      XVisualInfo *vi;
-	      int visual_depth;
-	      vi = glXGetVisualFromFBConfig(dpy, fbConfigs[j]);
-	      visual_depth = vi->depth;
-	      XFree(vi);
-	      if (visual_depth != i)
+	    XVisualInfo *vi;
+	    int		visualDepth;
+
+	    vi = glXGetVisualFromFBConfig (dpy, fbConfigs[j]);
+	    if (vi == NULL)
+		continue;
+
+	    visualDepth = vi->depth;
+
+	    XFree (vi);
+
+	    if (visualDepth != i)
 		continue;
-	    }
 
 	    (*s->getFBConfigAttrib) (dpy,
 				     fbConfigs[j],
@@ -1464,6 +1632,8 @@ addScreen (CompDisplay *display,
     s->redrawTime = 1000 / defaultRefreshRate;
     s->optimalRedrawTime = s->redrawTime;
 
+    reshape (s, s->attrib.width, s->attrib.height);
+
     glLightModelfv (GL_LIGHT_MODEL_AMBIENT, globalAmbient);
 
     glEnable (GL_LIGHT0);
@@ -1478,33 +1648,6 @@ addScreen (CompDisplay *display,
     s->lighting	      = FALSE;
     s->slowAnimations = FALSE;
 
-    if ( !XineramaIsActive( display->display ) )
-    {
-        s->useXinerama = FALSE;
-        s->headsInfo = 0;
-        s->headsCount = 0;
-    }
-    else
-    {
-        s->useXinerama = TRUE;
-
-        XineramaScreenInfo *screen_info;
-        int number;
-        screen_info = XineramaQueryScreens ( display->display, &number );
-        s->headsInfo = malloc ( sizeof (CompScreenHead) * number );
-        s->headsCount = number;
-        for ( i = 0; i < number; i++ )
-        {
-            s->headsInfo[i].areaX = s->headsInfo[i].x = screen_info[i].x_org;
-            s->headsInfo[i].areaY = s->headsInfo[i].y = screen_info[i].y_org;
-            s->headsInfo[i].areaWidth = s->headsInfo[i].width = screen_info[i].width;
-            s->headsInfo[i].areaHeight = s->headsInfo[i].height = screen_info[i].height;
-        }
-        XFree( screen_info );
-    }
-
-    reshape (s, s->attrib.width, s->attrib.height);
-
     addScreenToDisplay (display, s);
 
     screenInitPlugins (s);
@@ -1581,161 +1724,6 @@ addScreen (CompDisplay *display,
     return TRUE;
 }
 
-int 
-screenGetHead ( CompScreen *s,
-                int x,
-                int y,
-                int w,
-                int h )
-{
-    if ( !s->useXinerama ) return 1;
-
-    int i;
-    for ( i = 0; i < s->headsCount; i++ )
-    {
-        if ( x >= s->headsInfo[i].x &&
-             x < ( s->headsInfo[i].x + s->headsInfo[i].width ) &&
-             y >= s->headsInfo[i].y &&
-             y < ( s->headsInfo[i].y + s->headsInfo[i].height ) )
-            return i + 1;
-    }
-
-    if ( w != 0 && h != 0 )
-    {
-        int bestHead, maxArea, currentArea;
-        int windowLeftX, windowRightX, windowTopY, windowBottomY;
-        int areaLeftX, areaRightX, areaTopY, areaBottomY;
-        windowLeftX = x - w / 2;
-        windowRightX = x + w / 2;
-        windowTopY = y - h / 2;
-        windowBottomY = y + h / 2;
-        bestHead = maxArea = currentArea = 0;
-        for ( i = 0; i < s->headsCount; i++ )
-        {
-            if ( windowLeftX > ( s->headsInfo[i].x + s->headsInfo[i].width ) ||
-                 windowTopY > ( s->headsInfo[i].y + s->headsInfo[i].height ) ||
-                 windowRightX < s->headsInfo[i].x ||
-                 windowBottomY < s->headsInfo[i].y )
-                continue;
-            if ( windowLeftX <= s->headsInfo[i].x ) areaLeftX = s->headsInfo[i].x;
-            else areaLeftX = windowLeftX;
-            if ( windowRightX >= ( s->headsInfo[i].x + s->headsInfo[i].width ) )
-                areaRightX = s->headsInfo[i].x + s->headsInfo[i].width;
-            else areaRightX = windowRightX;
-            if ( windowTopY <= s->headsInfo[i].y ) areaTopY = s->headsInfo[i].y;
-            else areaTopY = windowTopY;
-            if ( windowBottomY >= ( s->headsInfo[i].y + s->headsInfo[i].height ) )
-                areaBottomY = s->headsInfo[i].y + s->headsInfo[i].height;
-            else areaBottomY = windowBottomY;
-            currentArea = ( areaRightX - areaLeftX ) * ( areaBottomY - areaTopY );
-            if ( currentArea > maxArea )
-            {
-                maxArea = currentArea;
-                bestHead = i + 1;
-            }
-        }
-        if ( bestHead > 0 )
-            return bestHead;
-    }
-    return s->headsCount + 1;
-}
-
-int
-screenGetHeadForWindow ( CompWindow *w )
-{
-    if ( !w->screen->useXinerama ) return 1;
-    return screenGetHead ( w->screen, w->serverX + w->width / 2, w->serverY + w->height / 2, w->width, w->height );
-}
-
-int
-screenGetCurrentHead ( CompScreen *s )
-{
-    if ( !s->useXinerama ) return 1;
-    int root_x = 0, root_y = 0;
-    int ignore_i;
-    unsigned int ignore_ui;
-
-    Window ignore_w;
-
-    XQueryPointer( s->display->display,
-                   s->root, &ignore_w, 
-                   &ignore_w, &root_x, &root_y,
-                   &ignore_i, &ignore_i, &ignore_ui );
-    return screenGetHead ( s, root_x, root_y, 0, 0 );
-}
-
-int
-screenGetHeadX ( CompScreen *s,
-                 int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return 0;
-    return s->headsInfo[ head - 1 ].x;
-}
-
-int
-screenGetHeadY ( CompScreen *s,
-                 int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return 0;
-    return s->headsInfo[ head - 1 ].y;
-}
-
-int
-screenGetHeadWidth ( CompScreen *s,
-                     int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->width;
-    return s->headsInfo[ head - 1 ].width;
-}
-
-int
-screenGetHeadHeight ( CompScreen *s,
-                      int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->height;
-    return s->headsInfo[ head - 1 ].height;
-}
-
-int
-screenGetHeadAreaX ( CompScreen *s,
-        int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->workArea.x;
-    return s->headsInfo[ head - 1 ].areaX;
-}
-
-int
-screenGetHeadAreaY ( CompScreen *s,
-        int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->workArea.y;
-    return s->headsInfo[ head - 1 ].areaY;
-}
-
-int
-screenGetHeadAreaWidth ( CompScreen *s,
-        int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->workArea.width;
-    return s->headsInfo[ head - 1 ].areaWidth;
-}
-
-int
-screenGetHeadAreaHeight ( CompScreen *s,
-        int head )
-{
-    if ( !s->useXinerama || head > s->headsCount )
-        return s->workArea.height;
-    return s->headsInfo[ head - 1 ].areaHeight;
-}
-
 void
 damageScreenRegion (CompScreen *screen,
 		    Region     region)
@@ -1792,27 +1780,6 @@ findWindowAtScreen (CompScreen *s,
     return 0;
 }
 
-Bool
-pointerOnlyOnDesktop (CompScreen *s,
-                      int pointerX,
-                      int pointerY)
-{
-    CompWindow *w;
-
-    if ( pointerX < s->workArea.x || pointerX > ( s->workArea.x + s->workArea.width ) || pointerY < s->workArea.y || pointerY > ( s->workArea.y + s->workArea.height ) )
-        return FALSE;
-
-	for (w = s->windows; w; w = w->next)
-	{
-        if (w->invisible || w->id == s->root || w->type & CompWindowTypeDesktopMask)
-            continue;
-        if ( pointerX >= w->attrib.x && pointerX <= ( w->attrib.x + w->attrib.width ) && pointerY >= w->attrib.y && pointerY <= ( w->attrib.y + w->attrib.height ) )
-            return FALSE;
-    }
-
-    return TRUE;
-}
-
 CompWindow *
 findTopLevelWindowAtScreen (CompScreen *s,
 			    Window      id)
@@ -2394,89 +2361,36 @@ void
 updateWorkareaForScreen (CompScreen *s)
 {
     CompWindow *w;
-		CompScreenHead *h;
-    int	       leftStrut, rightStrut, topStrut, bottomStrut, head;
+    int	       leftStrut, rightStrut, topStrut, bottomStrut;
     XRectangle workArea;
 
     leftStrut   = 0;
     rightStrut  = 0;
     topStrut    = 0;
     bottomStrut = 0;
-    head = 0;
 
-    if ( !s->useXinerama )
+    for (w = s->windows; w; w = w->next)
     {
-        for (w = s->windows; w; w = w->next)
-        {
-            if (w->attrib.map_state == IsUnmapped)
-                continue;
+	if (w->attrib.map_state == IsUnmapped)
+	    continue;
 
-            if (w->struts)
-	    {
-	        if (w->struts->left.width > leftStrut)
-		    leftStrut = w->struts->left.width;
+	if (w->struts)
+	{
+	    if (w->struts->left.width > leftStrut)
+		leftStrut = w->struts->left.width;
 
-                if (w->struts->right.width > rightStrut)
-                    rightStrut = w->struts->right.width;
+	    if (w->struts->right.width > rightStrut)
+		rightStrut = w->struts->right.width;
 
-                if (w->struts->top.height > topStrut)
-                    topStrut = w->struts->top.height;
+	    if (w->struts->top.height > topStrut)
+		topStrut = w->struts->top.height;
 
-                if (w->struts->bottom.height > bottomStrut)
-                    bottomStrut = w->struts->bottom.height;
-            }
-        }
-    }
-    else
-    {
-        // reset workarea for each head
-	for ( head = 0; head < s->headsCount; head++ )
-        {
-	    h = &s->headsInfo[ head ];
-            h->areaX = h->x;
-            h->areaY = h->y;
-	    h->areaWidth = h->width;
-	    h->areaHeight = h->height;
+	    if (w->struts->bottom.height > bottomStrut)
+		bottomStrut = w->struts->bottom.height;
 	}
-	for (w = s->windows; w; w = w->next)
-        {
-	    if (w->attrib.map_state == IsUnmapped)
-                continue;
-
-            if (w->struts)
-	    {
-		head = screenGetHeadForWindow(w);
-
-		h = &s->headsInfo[ head - 1 ];
-
-		if (w->struts->left.width > h->areaX)
-		    h->areaX = h->x + w->struts->left.width;
-
-		if (w->struts->top.height > h->areaY)
-		    h->areaY = h->y + w->struts->top.height;
-
-		if (w->struts->right.width > h->width - h->areaWidth - h->areaX) 
-		    h->areaWidth = h->width - (h->areaX - h->x) - w->struts->right.width;
-
-		if (w->struts->bottom.height > h->height - h->areaHeight - h->areaY)
-		    h->areaHeight = h->height - (h->areaY - h->y) - w->struts->bottom.height;
-
-		if (w->struts->left.width > leftStrut)
-		    leftStrut = w->struts->left.width;
-
-		if (w->struts->right.width > rightStrut)
-		    rightStrut = w->struts->right.width;
-
-		if (w->struts->top.height > topStrut)
-		    topStrut = w->struts->top.height;
-
-		if (w->struts->bottom.height > bottomStrut)
-		    bottomStrut = w->struts->bottom.height;
-            }
-        }
     }
 
-    #define MIN_SANE_AREA 100
+#define MIN_SANE_AREA 100
 
     if ((leftStrut + rightStrut) > (s->width - MIN_SANE_AREA))
     {
@@ -2727,32 +2641,50 @@ runCommand (CompScreen *s,
 void
 moveScreenViewport (CompScreen *s,
 		    int	       tx,
+		    int	       ty,
 		    Bool       sync)
 {
+    CompWindow *w;
+    int         m, wx, wy, vWidth, vHeight;
+
     tx = s->x - tx;
-    tx = MOD (tx, s->size);
+    tx = MOD (tx, s->hsize);
     tx -= s->x;
 
-    if (tx)
+    ty = s->y - ty;
+    ty = MOD (ty, s->vsize);
+    ty -= s->y;
+
+    if (!tx && !ty)
+	return;
+
+    s->x += tx;
+    s->y += ty;
+
+    tx *= -s->width;
+    ty *= -s->height;
+
+    vWidth = s->width * s->hsize;
+    vHeight = s->height * s->vsize;
+
+    for (w = s->windows; w; w = w->next)
     {
-	CompWindow *w;
-	int	   m, wx, vWidth = s->width * s->size;
+	if (w->attrib.override_redirect)
+	    continue;
 
-	s->x += tx;
+	if (w->type & (CompWindowTypeDesktopMask | CompWindowTypeDockMask))
+	    continue;
 
-	tx *= -s->width;
+	if (w->state & CompWindowStateStickyMask)
+	    continue;
 
-	for (w = s->windows; w; w = w->next)
+	/* x */
+	if (s->hsize == 1)
+	{
+	    wx = tx;
+	}
+	else
 	{
-	    if (w->attrib.override_redirect)
-		continue;
-
-	    if (w->type & (CompWindowTypeDesktopMask | CompWindowTypeDockMask))
-		continue;
-
-	    if (w->state & CompWindowStateStickyMask)
-		continue;
-
 	    m = w->attrib.x + tx;
 	    if (m - w->output.left < s->width - vWidth)
 		wx = tx + vWidth;
@@ -2760,28 +2692,48 @@ moveScreenViewport (CompScreen *s,
 		wx = tx - vWidth;
 	    else
 		wx = tx;
-
-	    if (w->saveMask & CWX)
-		w->saveWc.x += wx;
-
-	    moveWindow (w, wx, 0, sync, TRUE);
-
-	    if (sync)
-		syncWindowPosition (w);
 	}
 
+	if (w->saveMask & CWX)
+	    w->saveWc.x += wx;
+
+	/* y */
+	if (s->vsize == 1)
+	{
+	    wy = ty;
+	}
+	else
+	{
+	    m = w->attrib.y + ty;
+	    if (m - w->output.top < s->height - vHeight)
+		wy = ty + vHeight;
+	    else if (m + w->height + w->output.bottom > vHeight)
+		wy = ty - vHeight;
+	    else
+		wy = ty;
+	}
+
+	if (w->saveMask & CWY)
+	    w->saveWc.y += wy;
+
+	/* move */
+	moveWindow (w, wx, wy, sync, TRUE);
+
 	if (sync)
-	{
-	    unsigned long data[2];
+	    syncWindowPosition (w);
+    }
 
-	    data[0] = s->x * s->width;
-	    data[1] = 0;
+    if (sync)
+    {
+	unsigned long data[2];
 
-	    XChangeProperty (s->display->display, s->root,
-			     s->display->desktopViewportAtom,
-			     XA_CARDINAL, 32, PropModeReplace,
-			     (unsigned char *) data, 2);
-	}
+	data[0] = s->x * s->width;
+	data[1] = s->y * s->height;
+
+	XChangeProperty (s->display->display, s->root,
+			 s->display->desktopViewportAtom,
+			 XA_CARDINAL, 32, PropModeReplace,
+			 (unsigned char *) data, 2);
     }
 }
 
@@ -2790,7 +2742,7 @@ moveWindowToViewportPosition (CompWindow *w,
 			      int	 x,
 			      Bool       sync)
 {
-    int	tx, vWidth = w->screen->width * w->screen->size;
+    int	tx, vWidth = w->screen->width * w->screen->hsize;
 
     x += w->screen->x * w->screen->width;
     x = MOD (x, vWidth);
@@ -2918,7 +2870,10 @@ applyStartupProperties (CompScreen *screen,
     }
 
     if (s)
-	window->initialViewport = s->viewport;
+    {
+	window->initialViewportX = s->viewportX;
+	window->initialViewportY = s->viewportY;
+    }
 }
 
 void
@@ -3027,32 +2982,6 @@ sendWindowActivationRequest (CompScreen *s,
 		&xev);
 }
 
-void
-changeToWindowViewport (CompScreen *s,
-			     Window	id, CompWindow *w)
-{
-    XEvent xev;
-
-    xev.xclient.type    = ClientMessage;
-    xev.xclient.display = s->display->display;
-    xev.xclient.format  = 32;
-
-    xev.xclient.message_type = s->display->desktopViewportAtom;
-    xev.xclient.window	     = s->root;
-
-    xev.xclient.data.l[0] = defaultViewportForWindow (w) * s->width;
-    xev.xclient.data.l[1] = 0;
-    xev.xclient.data.l[2] = 0;
-    xev.xclient.data.l[3] = 0;
-    xev.xclient.data.l[4] = 0;
-
-    XSendEvent (s->display->display,
-		s->root,
-		FALSE,
-		SubstructureRedirectMask | SubstructureNotifyMask,
-		&xev);
-}
-
 void
 screenTexEnvMode (CompScreen *s,
 		  GLenum     mode)
@@ -3127,7 +3056,7 @@ makeScreenCurrent (CompScreen *s)
 {
     if (currentRoot != s->root)
     {
-	glXMakeCurrent (s->display->display, s->root, s->ctx);
+	glXMakeCurrent (s->display->display, s->output, s->ctx);
 	currentRoot = s->root;
 
 	s->pendingCommands = TRUE;
@@ -3142,6 +3071,27 @@ finishScreenDrawing (CompScreen *s)
 	makeScreenCurrent (s);
 	glFinish ();
 
-	//s->pendingCommands = FALSE;
+	s->pendingCommands = FALSE;
     }
 }
+
+int
+outputDeviceForPoint (CompScreen *s,
+		      int	 x,
+		      int	 y)
+{
+    int i, x1, y1, x2, y2;
+
+    for (i = 0; i < s->nOutputDev; i++)
+    {
+	x1 = s->outputDev[i].region.extents.x1;
+	y1 = s->outputDev[i].region.extents.y1;
+	x2 = s->outputDev[i].region.extents.x2;
+	y2 = s->outputDev[i].region.extents.y2;
+
+	if (x1 < x && x2 >= x && y1 < y && y2 >= y)
+	    return i;
+    }
+
+    return s->currentOutputDev;
+}
